import * as WebSocket from 'ws';
import ClientManager from '../clientManager';
import ChannelManager from '../channelManager';
import PublicChannel from '../channels/types/publicChannel';
import PrivateChannel from '../channels/types/privateChannel';
import CustomChannel from '../channels/types/customChannel';

const messageManager = require('../messageManager');
const logger = require('../logger');

class ClientBase {
  ws: WebSocket;
  data: any;
  id: number;
  channel: PublicChannel|PrivateChannel|CustomChannel|null;
  clientManager: ClientManager;
  channelManager: ChannelManager;
  roles: Array<string>;
  messageListener: (data: any) => void;

  constructor(data: any, ws: WebSocket, channelManager: ChannelManager, clientManager: ClientManager) {
    this.ws = ws;
    this.data = data;
    this.id = data.user_id;
    this.channel = null;
    this.clientManager = clientManager;
    this.channelManager = channelManager;
    this.roles = ['receiver']
    this.messageListener = (data: any) => {
      if (this.channel) {
        logger.accessLog.info(`starting message transaction on channel ${this.channel.id}: `, {data});
        data = messageManager.prepareMessage(data, this.channel, this);
        this.messageTransaction(data);
      }
    };
  }

  getData() {
    return this.data;
  }

  type() {
    return this.data.client_type;
  }

  clientType() {
    return this.data.client;
  }

  connectToChannel(channel: PublicChannel|PrivateChannel|CustomChannel) {
    this.channel = channel;

    this.ws.on('message', this.messageListener);

    this.ws.on('close', (reasonCode: string, description: string) => {
      logger.accessLog.info(`closing connection for client ${this.id}`);

      if (this.channel) {
        this.channel.removeClient(this.id);
      }

      this.clientManager.removeClient(this.id);
      logger.accessLog.info(`closed connection for client ${this.id}`);
    });
  }

  messageTransaction(data: any) {
    if (this.channel && !data.error) {
      switch (data.message_type) {
        case 'broadcast':
          this.channel.broadcastMessage(this, data);
          break;
        case 'direct':
          let to = this.clientManager.getClient(data.message.to)
          to.directMessage(data);
          break;
        case 'changeChannel':
          this.ws.removeListener('message', this.messageListener);
          this.channelManager.changeChannel(this, data);
          break;
      }

      logger.accessLog.info(`message transaction complete on channel ${this.channel.id}: `, {message: data});
    } else {
      logger.errorLog.info(`Validation failed or client is not part of a channel, please review schema`, {data: {message: data, error: data.error}});
    }
  }

  directMessage(message: any) {
    this.ws.send(JSON.stringify(message));
    logger.accessLog.info(`sent direct message to ${this.id}`, { data: { message }});
  }

  replaceWebSocket(ws: WebSocket) {
    this.ws.close();
    this.ws = ws;
  }
}

export default ClientBase;
